# pyCalibration automated tests

## Objective

Test available detector calibrations against picked reference data
to prove the software consistency and avoid sneaky bugs that
affect the produced data quality.

1. Test the successful processing for executing SLURM jobs.
   1. xfel-calibrate CL runs successfully
   2. SLURM jobs are executed on Maxwell
   3. SLURM jobs are COMPLETED at the end.
2. Test the availability of files in the out-folder
   1. Validate the presence of a PDF report.
   2. Validate the number of HDF5 files against the number of HDF5 files in the reference folder.
3. Validate the numerical values of the processed data against the referenced data.


These tests are meant to run on all detector calibrations before any release. As well as run it per branch on the selected detector/calibration that is affected by the branch's changes.

## Current state

- Tests are defined by a callab_test.py DICT.
  ```py
  {
   "<test-name>(`<detector-calibration-operationMode>`)": {
      "det-type": "<detectorType>",
      "cal_type": "<calibrationType>",
      "config": {
         "in-folder": "<inFolderPath>",
         "out-folder": "<outFolderPath>",
         "run": "<runNumber>",
         "karabo-id": "detector",

         "<config-name>": "<config-value>",
         ...: ...,
      },
      "reference-folder": "{}/{}/{}",
   }
   ...
  }
  ```

- Tests are triggered using CLI:
  - `pytest tests/test_reference_runs/test_pre_deployment.py --release-test --reference-folder <reference-folder-path> --out-folder <out-folder-path>`
  - Arguments:
    - required arguments:
      - release-test: this is needed to trigger the automated test. To avoid triggering this as a part of the Gitlab CI this boolean was created.
      - reference-folder: Setting the reference folder path. The reference folder is expected to have exactly the same structure as the out-folder. Usually the reference folders are out-folder from previous successful tested releases.
      - out-folder: The output folder paths for saving the test output files.
        - The structure is `<detector>/<test-name>/[PDF, HDF5, YAML, SlurmOut, ...]`
    - optional arguments:
      - picked-test: this can be used to only run the tests for one <test-name> only.
      - calibration: this can be used to pick only one calibration type to run the test for. [dark or correct]
      - detector: this can be used to pick detectors to run the test for and skip the rest.
      - no-numerical-validation: as the numerical validation is done by default. This argument can be used to skip it and stop at executing the calibration and making sure that the SLURM jobs were COMPLETED.
      - validation-only: in case the test output and reference files were already available and only a validation check is needed. this argument can be used to only run the validation checks without executing any calibration jobs.
      - find-difference: by default files are compared to reference data. If more information is needed to know the difference between reference and test files, this argument can be helpful. It should be expected that more time is consumed to check the files that fails the validation, to get a report for the different datasets/attributes.

- Below are the steps taken to fully test the calibration files:
   1. Run `xfel-calibrate DET CAL ...`, this will result in Slurm calibration jobs.
   2. Check the Slurm jobs state after they finish processing.
   3. Confirm that there is a PDF available in the output folder.
   4. Validate the HDF5 files.
      1. Compare the MD5 checksum for the output and reference.
      2. Find the datatsets/attributes that are different in both files.

In case a test fails the whole test fails and the next test starts.

## Challenge for manually triggering the automated tests.

Currently these tests run by only one person before each release. Unfortunately, it is not use on individual branches and MRs as it is not exposed to all contributors.

Automating this test can solve the dependence on only one user to run the tests manually. Additionally, it can help in identifying bugs/errors faster, instead of delaying the tests to one duration before the release. This can as well help in not delaying the releases by some hours as the full test can take a lot of hours.

It is preferred to run this automated test in a common place that is accessed by the calibration team to have more operators/monitors. This would result in a more active approach in solving any problem related to the testing pipeline or the tested detector calibrations.
Moreover exposing the running test would help in collecting more ideas and efforts in improving the testing pipeline.

## Automating triggering the testing pipeline.

To automate the triggering:
  - Decide on a max-node to run the tests.
    - Currently the tests are manually triggered on a selected node and after the calibration execution is done all checks and validations are done on this selected node.
    - Create a cron job to schedule the automatic triggering for the tests.
    - Keep a logging format as a reporting tool for the test results.
